import asyncio
import logging
import shlex
import html
from prompt_toolkit.formatted_text import HTML

from .StreamClient import StreamClient

_log = logging.getLogger("katcpcli.CommandDispatcher")


class CommandDispatcher:
    """
    Dispatch a command-line to the appropriate method.
    """

    def __init__(self, output_stream, log_stream):
        self.commands = {
            "connect": self.connect,
            "reconnect": self.reconnect,
            "log_inform": self.log_inform,
        }
        self.log_stream = log_stream
        self.output_stream = output_stream
        self.katcpclient = None
        self.completion_set = set(self.commands.keys())
        self.supressed_logs = 0
        self.running_tasks = set()

    async def __call__(self, line):
        """
        Act on a prompt line.
        """
        try:
            line = shlex.split(line)
        except ValueError as E:
            self.output_stream(
                HTML(
                    f' <style fg="ansired"> Error</style> processing line:<b> {html.escape(line)}</b> - {E}'
                )
            )
            return

        if not line:
            return
        cmd = line[0]
        args = line[1:]
        if cmd.startswith("?"):
            cmd = "katcpcmd"
            args.insert(0, line[0][1:])

        try:
            task = asyncio.create_task(self.commands[cmd](*args))
            self.running_tasks.add(task)
            task.add_done_callback(self.running_tasks.discard)
        except KeyError:
            self.output_stream(
                HTML(
                    f' <style fg="ansired"> Error</style> Unknown command:<b> {html.escape(cmd)}</b>'
                )
            )
        except TypeError as E:
            self.output_stream(E)

    async def update_command_completion(self):
        """
        Update command completion set
        """
        # remove requests from completion list
        to_remove = set()
        for s in self.completion_set:
            if s.startswith("?"):
                to_remove.add(s)
        self.completion_set -= to_remove

        requests = await self.katcpclient.get_requests()
        for s in requests:
            self.completion_set.add(f"?{s}")

    async def connect(self, host, port=None):
        """
        Connect to host:port
        """
        if port is None:
            try:
                host, port = host.split(":")
            except ValueError:
                self.output_stream(
                    HTML(
                        '<style fg="ansired"> Error</style> Expecting host:port or host port'
                    )
                )
                return

        if self.katcpclient:
            self.output_stream(
                f"Closing connection to {self.katcpclient.host}:{self.katcpclient.port}"
            )
            self.katcpclient.close()
            await self.katcpclient.wait_closed()
            self.katcpclient = None
        self.output_stream(f"Connecting to {host}:{port} ...")
        self.katcpclient = StreamClient(
            self.output_stream, self.log_stream, host, port
        )
        await self.katcpclient.wait_connected()
        await self.update_command_completion()
        self.output_stream(f"Connection to {host}:{port} established.")

        self.commands["katcpcmd"] = self.katcpclient.handle_request
        self.supressed_logs = 0

    async def reconnect(self):
        """
        Reconnect to the previous host
        """
        if self.katcpclient:
            self.output_stream("Reconnecting")
            await self.connect(self.katcpclient.host, self.katcpclient.port)
        else:
            self.output_stream("No connection to reconnect to!")

    async def log_inform(self, args):
        """
        Toggle log inform handling.
        """
        if args == "enabled":
            self.katcpclient.log_stream = self.log_stream
        elif args == "disabled":
            self.katcpclient.log_stream = self._log_counter
        else:
            self.output_stream("Requires one of [enabled/disabled] as arguement")
            return

        self.output_stream(f"Showing log informs is now {args}")

    def _log_counter(self, line):  # pylint: disable=unused-argument
        self.supressed_logs += 1



import logging
import time
import datetime
import html

import aiokatcp
from prompt_toolkit.formatted_text import HTML


_log = logging.getLogger("katcpcli.StreamClient")


def sensor_status_format(value):
    """
    Format sensor status value.
    """
    name = aiokatcp.Sensor.Status(int(value)).name

    colors = {
        "UNKNOWN": "ansiblue",
        "NOMINAL": "ansigreen",
        "WARN": "ansiyellow",
        "ERROR": "ansired",
        "FAILURE": "ansired",
        "UNREACHABLE": "anisred",
        "INACTIVE": "cyan",
    }

    col = colors.get(name)
    if col:
        res = f'<style fg="{col}">{name}</style>'
    else:
        res = name
    _log.debug("Formatted status name: %s", res)
    return res


class StreamClient(aiokatcp.client.Client):
    """
    KATCP client that creates output lines from the replies and writes them to
    different streams for log messages and default output.
    """

    def __init__(self, output_stream, log_stream, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.output_stream = output_stream
        self.log_stream = log_stream

    def _decode(self, values):
        return [html.escape(v.decode("ascii")) for v in values]

    def unhandled_inform(self, msg: aiokatcp.core.Message) -> None:
        """Deal with unhandled informs"""
        args = " ".join(self._decode(msg.arguments))
        self.output_stream(HTML(f"unhandled-inform:\n#{msg.name} {args}"))

    def inform_client_connected(self, *args):
        """Deal with connect informs"""
        msg = " ".join(self._decode(args))
        self.log_stream(HTML(f"{time.ctime()} Client connected: {msg}"))

    def inform_client_disconnected(self, *args):
        """Deal with disconnect informs"""
        msg = " ".join(self._decode(args))
        self.log_stream(HTML(f"{time.ctime()} Client disconnected: {msg}"))

    def inform_log(self, *args):
        """Deal with connect informs"""
        msg = " ".join(self._decode(args))
        self.log_stream(HTML(f"{msg}"))

    def _format_sensor_list(self, *args):
        """Deal with sensor values"""
        arguments = self._decode(args)
        if len(arguments) >= 5:
            if arguments[4].startswith("iVBORw0KGgo"):
                data = "Some base64 encoded png image (hidden)"
            else:
                data = arguments[4]

            output = "{} {}  {:30}   {}".format(
                datetime.datetime.fromtimestamp(float(arguments[0])).isoformat(),
                sensor_status_format(arguments[1]),
                arguments[2],
                data,
            )
        else:
            output = " ".join(arguments)
        return output.rstrip()
        self.output_stream(HTML(output))

    def inform_sensor_value(self, *args):
        """Deal with sensor values"""
        output = self._format_sensor_list(*args)
        self.output_stream(HTML(output))

    def inform_sensor_status(self, *args):
        """Deal with sensor values"""
        output = self._format_sensor_list(*args)
        self.log_stream(HTML(output))

    def inform_sensor_list(self, *args):
        """Deal with sensor list"""
        if len(args) > 1:
            R = "; ".join(self._decode(args[1:]))
        else:
            R = ""
        output = HTML(" - {}\n              {}".format(self._decode((args[0],))[0], R))
        self.output_stream(output)

    def inform_help(self, *args):
        """Pretty print help"""
        if len(args) > 1:
            R = "; ".join(self._decode(args[1:]))
        else:
            R = ""
        output = HTML(" - {}\n              {}".format(self._decode((args[0],))[0], R))
        self.output_stream(output)

    def print_command_reply(self, cmd, rv, res=""):
        """
        Print reply of a command.
        """
        if rv == "ok":
            col = "ansigreen"
        else:
            col = "ansired"
        self.output_stream(HTML(f'<b>!{cmd}</b> <style fg="{col}">{rv}</style> {res}'))

    async def handle_request(self, *args):
        """
        handle request reply
        """
        _log.debug("handle request called with args %s", args)
        if not args:
            return
        cmd = args[0]
        try:
            reply, informs = await self.request(cmd, *args[1:])
            for inform in informs:
                self.handle_inform(inform)
            rv = " ".join(self._decode(reply))
            self.print_command_reply(cmd, "ok", rv)
        except aiokatcp.connection.FailReply as E:
            self.print_command_reply(cmd, "fail", E)
        except aiokatcp.connection.InvalidReply as E:
            self.print_command_reply(cmd, "invalid", E)
        except BrokenPipeError:
            self.print_command_reply(cmd, "not connected!")
        except ConnectionResetError:
            self.print_command_reply(cmd, "connection to server lost!")
        except Exception as E:  # pylint: disable=broad-except
            _log.error("Unhandled exeption during request handling")
            _log.exception(E)
            self.print_command_reply(cmd, "Unexpected exception during handling of command!")

    async def get_requests(self):
        """
        Get requests for device
        """
        requests = set()
        try:
            _, informs = await self.request("help")
            for inform in informs:
                requests.add(inform.arguments[0].decode("ascii"))
        except Exception as E:  # pylint: disable=broad-except
            _log.error("Error during command retrival")
            _log.exception(E)
        return requests

    def _warn_failed_connect(self, exc: Exception) -> None:
        self.logger.debug("Failed to connect to %s:%s: %s", self.host, self.port, exc)
